"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const utils_1 = require("./utils");
/**
 *
 * @param node
 * @param symbol
 * @param sourceFile
 * @param priorDiagnostics
 * @param log
 * @param declarationOverride  - allows us to override the declaration of the symbol if the deprecation was found elsewhere
 * @returns
 */
const createDeprecatedDiagnostic = (node, symbol, sourceFile, priorDiagnostics, log, declarationOverride) => {
    log(`Creating diagnostic for deprecated symbol: ${symbol.getName()} - ${node.getText()}`);
    const diagnosticStart = node.getStart();
    const diagnosticEnd = node.getEnd() - node.getStart();
    const diagnosticCode = 6385; // Deprecated code
    const declaration = declarationOverride || symbol.getDeclarations()[0];
    const deprecationSourceStart = declaration.getStart();
    const deprecationSourceEnd = declaration.getEnd();
    const deprecationSourceFile = declaration.getSourceFile();
    const alreadyReported = priorDiagnostics.some(diagnostic => diagnostic.start === diagnosticStart &&
        diagnostic.length === diagnosticEnd &&
        diagnostic.code === diagnosticCode);
    if (!alreadyReported) {
        log(`Adding diagnostic for deprecated symbol: ${symbol.getName()}`);
        return {
            file: sourceFile,
            start: diagnosticStart,
            length: diagnosticEnd,
            messageText: `'${symbol.getName()}' is deprecated.`,
            category: ts.DiagnosticCategory.Warning,
            reportsDeprecated: true,
            code: diagnosticCode,
            relatedInformation: [
                {
                    start: deprecationSourceStart,
                    length: deprecationSourceEnd - deprecationSourceStart,
                    file: deprecationSourceFile,
                    messageText: "The declaration was marked as deprecated here.",
                    category: ts.DiagnosticCategory.Error,
                    code: 2798
                }
            ]
        };
    }
    return null;
};
// Helper function to create detailed deprecation message for QuickInfo
const createDeprecatedQuickInfoTag = (symbol, checker) => {
    const declaration = symbol.getDeclarations()[0];
    const sourceFile = declaration.getSourceFile();
    const start = declaration.getStart();
    const end = declaration.getEnd();
    return {
        name: "deprecated",
        text: [
            {
                kind: "text",
                text: `Symbol '${symbol.getName()}' is deprecated.`
            },
            {
                kind: "space",
                text: " "
            },
            {
                kind: "text",
                text: `Declared in ${sourceFile.fileName} at line ${sourceFile.getLineAndCharacterOfPosition(start).line + 1}.`
            }
        ]
    };
};
const exportSymbolCache = new Map();
const wildCardExportCache = new Map();
const importCache = new Map();
const getImportCacheKey = (importDeclaration, targetSymbol) => {
    return (importDeclaration.getSourceFile().fileName + "-" + importDeclaration.getText() + "-" + targetSymbol.getName());
};
const getCachedDiagnosticForImportDeclaration = (importDeclaration, targetSymbol) => {
    const cacheKey = getImportCacheKey(importDeclaration, targetSymbol);
    return importCache.get(cacheKey);
};
const addCachedDiagnosticForImportDeclaration = (importDeclaration, targetSymbol, diagnostic) => {
    const cacheKey = getImportCacheKey(importDeclaration, targetSymbol);
    const existingCache = importCache.get(cacheKey);
    if (existingCache) {
        existingCache.diagnostics = existingCache.diagnostics.concat(diagnostic);
        existingCache.isDeprecated = true;
        importCache.set(cacheKey, existingCache);
    }
    else {
        importCache.set(cacheKey, {
            importDeclaration,
            symbol: targetSymbol,
            diagnostics: diagnostic,
            isDeprecated: true
        });
    }
};
// Traverse imports and re-exports to resolve and check for deprecations
const isImportDeclarationDeprecated = (currentNode, checker, log, program, optionalSymbolToMatch) => {
    var _a;
    const diagnostics = [];
    // Get the import node for the current node
    let node = currentNode;
    // Handle Import Declarations
    if (ts.isImportDeclaration(node)) {
        log(`[IMPORT] Checking import declaration: ${node.getText()}`);
        const importClause = node.importClause;
        if (importClause && importClause.namedBindings) {
            const namedBindings = importClause.namedBindings;
            if (ts.isNamedImports(namedBindings)) {
                const moduleSpecifier = node.moduleSpecifier;
                const moduleName = moduleSpecifier.text;
                const sourceFile = node.getSourceFile();
                log(`[IMPORT] Resolving module specifier: "${moduleName}" from "${sourceFile.fileName}"`);
                // Resolve the module to get the file name
                const resolvedModule = ts.resolveModuleName(moduleName, sourceFile.fileName, program.getCompilerOptions(), ts.sys);
                if (resolvedModule.resolvedModule) {
                    const resolvedFileName = resolvedModule.resolvedModule.resolvedFileName;
                    log(`[IMPORT] Resolved module "${moduleName}" to "${resolvedFileName}"`);
                    // Skip node_modules
                    if (resolvedFileName.includes("node_modules")) {
                        log(`[IMPORT] Skipping node_modules`);
                        return [];
                    }
                    const moduleSourceFile = program.getSourceFile(resolvedFileName);
                    if (moduleSourceFile) {
                        // Get the module symbol from the source file
                        const moduleSymbol = checker.getSymbolAtLocation(moduleSourceFile);
                        if (moduleSymbol) {
                            const exports = checker.getExportsOfModule(moduleSymbol);
                            log(`[IMPORT] Found ${exports.length} exports in module "${moduleSourceFile.getSourceFile().fileName}"`);
                            for (const element of namedBindings.elements) {
                                const importName = element.name.text;
                                const elementSymbol = checker.getSymbolAtLocation(element.name);
                                const cachedImportResult = getCachedDiagnosticForImportDeclaration(node, elementSymbol);
                                // if (cachedImportResult) {
                                //     log(`[IMPORT] Found cached import result for symbol: ${elementSymbol.getName()}`);
                                //     return cachedImportResult.diagnostics;
                                // }
                                let exportSymbol = exports.find(s => s.name === importName);
                                // if (exportSymbol && exportSymbolCache.has(exportSymbol)) {
                                //     log(`[IMPORT] Found cached export symbol: ${exportSymbol.getName()}`);
                                //     const exportSymbolDiagnostic = exportSymbolCache.get(exportSymbol);
                                //     if (exportSymbolDiagnostic) {
                                //         exportSymbolDiagnostic.start = currentNode.getStart();
                                //         exportSymbolDiagnostic.length = currentNode.getEnd() - currentNode.getStart();
                                //         diagnostics.push(exportSymbolDiagnostic);
                                //         addCachedDiagnosticForImportDeclaration(node, exportSymbol, [
                                //             exportSymbolDiagnostic
                                //         ]);
                                //     }
                                //     continue;
                                // }
                                if (!exportSymbol) {
                                    let wildcardExports = wildCardExportCache.get(moduleSymbol);
                                    if (!wildcardExports) {
                                        wildcardExports = [];
                                        const allExports = moduleSymbol.exports;
                                        allExports.forEach(exportEntry => {
                                            log(`[IMPORT] Export Entry: ${exportEntry.getName()}`);
                                            // Check if the import is a wildcard import
                                            log(`[IMPORT] Checking optional symbol: ${optionalSymbolToMatch === null || optionalSymbolToMatch === void 0 ? void 0 : optionalSymbolToMatch.getName()}`);
                                            // Filter to all exports where name == "__export"
                                            if (exportEntry.getName() === "__export") {
                                                wildcardExports.push(exportEntry);
                                            }
                                        });
                                        wildCardExportCache.set(moduleSymbol, wildcardExports);
                                    }
                                    log(`[IMPORT] Found ${wildcardExports.length} wildcard exports`);
                                    wildcardExports.forEach(wildcardExport => {
                                        var _a;
                                        const exportSymbolDeclaration = (_a = wildcardExport
                                            .declarations) === null || _a === void 0 ? void 0 : _a[0];
                                        log(`[IMPORT] Checking wildcard export declaration: ${exportSymbolDeclaration === null || exportSymbolDeclaration === void 0 ? void 0 : exportSymbolDeclaration.getText()}`);
                                        log(`[IMPORT] Wildcard module specifier: ${exportSymbolDeclaration.moduleSpecifier.getText()}`);
                                        const resolvedWildcardModule = ts.resolveModuleName(exportSymbolDeclaration.moduleSpecifier.text, sourceFile.fileName, program.getCompilerOptions(), ts.sys);
                                        log(`[IMPORT] Resolved wildcard module: ${resolvedWildcardModule.resolvedModule}`);
                                        const resolvedWildcardFileName = resolvedWildcardModule.resolvedModule.resolvedFileName;
                                        log(`[IMPORT] Resolved module "${exportSymbolDeclaration.moduleSpecifier.getText()}" to "${resolvedWildcardFileName}"`);
                                        const wildcardModuleSourceFile = program.getSourceFile(resolvedWildcardFileName);
                                        // Get the module symbol from the source file
                                        const wildcardModuleSymbol = checker.getSymbolAtLocation(wildcardModuleSourceFile);
                                        const wildcardModuleExports = checker.getExportsOfModule(wildcardModuleSymbol);
                                        log(`[IMPORT] Wildcard exports: ${wildcardModuleExports
                                            .map(e => e.getName())
                                            .join(", ")}`);
                                        /**
                                         * If optionalSymbolToMatch is provided we assume to know the symbol name we are looking for.
                                         * If using importName as the symbol name, we will check if the wildcard module exports contain the symbol name.
                                         */
                                        const importNameToMatch = (optionalSymbolToMatch === null || optionalSymbolToMatch === void 0 ? void 0 : optionalSymbolToMatch.getName()) || importName;
                                        const hasTargetExport = wildcardModuleExports.find(e => e.getName() === importNameToMatch);
                                        if (hasTargetExport) {
                                            exportSymbol = wildcardExport;
                                        }
                                    });
                                }
                                else {
                                    // If the export symbol name doesn't match the current symbol, skip
                                    if (optionalSymbolToMatch &&
                                        (exportSymbol === null || exportSymbol === void 0 ? void 0 : exportSymbol.getName()) !== optionalSymbolToMatch.getName()) {
                                        log(`[IMPORT] Skipping export "${importName} - ${exportSymbol === null || exportSymbol === void 0 ? void 0 : exportSymbol.getName()} - ${optionalSymbolToMatch.getName()}"`);
                                        continue;
                                    }
                                }
                                const exportSymbolDeclaration = (_a = exportSymbol === null || exportSymbol === void 0 ? void 0 : exportSymbol.getDeclarations()) === null || _a === void 0 ? void 0 : _a[0];
                                /**
                                 * Check if the parent node of the export symbol is deprecated.
                                 */
                                let exportSymbolParent = exportSymbolDeclaration === null || exportSymbolDeclaration === void 0 ? void 0 : exportSymbolDeclaration.parent;
                                log(`[IMPORT] Checking parent node for deprecation: ${exportSymbolParent === null || exportSymbolParent === void 0 ? void 0 : exportSymbolParent.getText()}`);
                                while (exportSymbolParent && !ts.isSourceFile(exportSymbolParent)) {
                                    const isExportSymbolParentDeprecated = (0, utils_1.isNodeDeprecated)(checker, exportSymbolParent);
                                    log(`[IMPORT] Parent node is deprecated: ${isExportSymbolParentDeprecated}`);
                                    if (isExportSymbolParentDeprecated) {
                                        let importNode = undefined;
                                        node.forEachChild(child => {
                                            log(`[IMPORT] Child: ${child.getText()}`);
                                            child.forEachChild(namedImport => {
                                                log(`[IMPORT] Named Import: ${namedImport.getText()}`);
                                                namedImport.forEachChild(importSpec => {
                                                    log(`[IMPORT] Import Specifier: ${importSpec.getText()}`);
                                                    if (importSpec.getText() === importName) {
                                                        importNode = importSpec;
                                                    }
                                                });
                                            });
                                        });
                                        const diagnostic = createDeprecatedDiagnostic(importNode, exportSymbol, exportSymbolDeclaration.getSourceFile(), [], log, exportSymbolDeclaration);
                                        diagnostics.push(diagnostic);
                                        addCachedDiagnosticForImportDeclaration(node, exportSymbol, [diagnostic]);
                                        exportSymbolCache.set(exportSymbol, diagnostic);
                                    }
                                    exportSymbolParent = exportSymbolParent === null || exportSymbolParent === void 0 ? void 0 : exportSymbolParent.parent;
                                }
                                if (exportSymbol) {
                                    const cachedImportResult = getCachedDiagnosticForImportDeclaration(node, exportSymbol);
                                    // if (cachedImportResult) {
                                    //     log(
                                    //         `[IMPORT] Found cached import result for symbol: ${exportSymbol.getName()}`
                                    //     );
                                    //     return cachedImportResult.diagnostics;
                                    // }
                                    log(`[IMPORT] Checking export "${importName}" in module "${moduleSourceFile.getSourceFile().fileName}"`);
                                    const declarations = exportSymbol.getDeclarations();
                                    if (declarations && declarations.length > 0) {
                                        const declaration = declarations[0];
                                        const declarationComments = (0, utils_1.getCommentsFromDeclaration)(declaration);
                                        if (declarationComments.length > 0) {
                                            log(`[IMPORT] Found comments on declaration: ${declarationComments.join("\n")}`);
                                        }
                                        if ((0, utils_1.isDeclarationDeprecated)(declaration)) {
                                            log(`[IMPORT] Import "${importName}" is deprecated in "${moduleName}"`);
                                            // Recursively search node for importSpecifier matching imoprtname
                                            const searchNodeForImportSpecifier = (node) => {
                                                if (ts.isImportSpecifier(node) && node.name.text === importName) {
                                                    return node;
                                                }
                                                return ts.forEachChild(node, searchNodeForImportSpecifier);
                                            };
                                            let importNode = searchNodeForImportSpecifier(node);
                                            log(`[IMPORT] Import Node: ${importNode === null || importNode === void 0 ? void 0 : importNode.getText()}`);
                                            const diagnostic = createDeprecatedDiagnostic(importNode, exportSymbol, node.getSourceFile(), [], log, declaration);
                                            diagnostics.push(diagnostic);
                                            addCachedDiagnosticForImportDeclaration(node, exportSymbol, [diagnostic]);
                                        }
                                    }
                                }
                                else {
                                    log(`[IMPORT] Could not find export "${importName}" in module "${moduleName}"`);
                                }
                            }
                        }
                        else {
                            log(`[IMPORT] Could not get symbol for module "${moduleName}"`);
                        }
                    }
                    else {
                        log(`[IMPORT] Could not get source file for "${resolvedFileName}"`);
                    }
                }
                else {
                    log(`[IMPORT] Could not resolve module "${moduleName}"`);
                }
            }
        }
        log(`[IMPORT] Returning diagnostics: ${diagnostics.length}`);
        return diagnostics;
    }
};
// The factory function that TypeScript expects
const init = ({ typescript: ts }) => {
    return {
        create: info => {
            var _a;
            const proxy = Object.create(null);
            const oldLanguageService = info.languageService;
            const oldGetQuickInfoAtPosition = oldLanguageService.getQuickInfoAtPosition;
            const oldGetSemanticDiagnostics = oldLanguageService.getSemanticDiagnostics;
            const oldGetSuggestionDiagnostics = oldLanguageService.getSuggestionDiagnostics;
            // Create a proxy to wrap the language service
            for (let k in oldLanguageService) {
                const x = oldLanguageService[k];
                proxy[k] = (...args) => x.apply(oldLanguageService, args);
            }
            // Logging to TypeScript Server
            const log = (message) => {
                // if (!message.includes("cached export") && !message.includes("Returning ")) {
                //     return;
                // }
                // Or use the logger if available
                if (info.project && info.project.projectService && info.project.projectService.logger) {
                    info.project.projectService.logger.info(`[DEPRECATION PLUGIN]: ${message}`);
                }
                else {
                    console.log(`[DEPRECATION PLUGIN]: ${message}`);
                }
            };
            log("Plugin Initialized - hotloading");
            const checker = (_a = info.languageService.getProgram()) === null || _a === void 0 ? void 0 : _a.getTypeChecker();
            // Common logic for both suggestion and semantic diagnostics
            const checkDiagnostics = (fileName, priorDiagnostics, category, program = info.languageService.getProgram()) => {
                var _a;
                const sourceFile = (_a = info.languageService.getProgram()) === null || _a === void 0 ? void 0 : _a.getSourceFile(fileName);
                const diagnostics = [];
                if (sourceFile && checker && (0, utils_1.isSupportedFileType)(fileName)) {
                    log(`Checking ${fileName} for deprecated symbols`);
                    const visit = (node) => {
                        var _a, _b;
                        // Check if node start/end/category/source file is already in diagnostics
                        diagnostics.some(diagnostic => {
                            if (diagnostic.start === node.getStart() &&
                                diagnostic.length === node.getEnd() - node.getStart() &&
                                diagnostic.category === category &&
                                diagnostic.file === sourceFile) {
                                return true;
                            }
                        });
                        // First we check for imports
                        const deprecatedAliasDiagnostics = isImportDeclarationDeprecated(node, checker, log, program);
                        if (deprecatedAliasDiagnostics === null || deprecatedAliasDiagnostics === void 0 ? void 0 : deprecatedAliasDiagnostics.length) {
                            for (const diagnostic of deprecatedAliasDiagnostics) {
                                diagnostic.category = category;
                                diagnostics.push(diagnostic);
                            }
                            return;
                        }
                        else if (!ts.isImportDeclaration(node) && !ts.isExportDeclaration(node)) {
                            // This will check for deprecated symbols in the source file by checkin the import declarations using the same logic as above
                            const symbol = checker.getSymbolAtLocation(node);
                            if (symbol) {
                                // Find the first declaration associated with the symbol
                                const declaration = (_a = symbol.getDeclarations()) === null || _a === void 0 ? void 0 : _a[0];
                                if (declaration && ts.isImportSpecifier(declaration)) {
                                    log(`Declaration found in import: ${declaration.getText()}`);
                                    let importDeclaration = declaration;
                                    // Get top-level import declaration
                                    while (importDeclaration.parent) {
                                        if (ts.isImportDeclaration(importDeclaration.parent)) {
                                            importDeclaration = importDeclaration.parent;
                                            break;
                                        }
                                        importDeclaration = importDeclaration.parent;
                                    }
                                    // Check if the import is deprecated
                                    if (ts.isImportDeclaration(importDeclaration)) {
                                        log(`Import parent: ${importDeclaration.getText()}`);
                                        // const cachedImportResult = getCachedDiagnosticForImportDeclaration(
                                        //     importDeclaration as ts.ImportDeclaration,
                                        //     symbol
                                        // );
                                        const cachedImportResult = undefined;
                                        if (cachedImportResult) {
                                            log(`Cached import for symbol: ${symbol.getName()}, found ${cachedImportResult.diagnostics.length} diagnostics`);
                                            if ((_b = cachedImportResult.diagnostics) === null || _b === void 0 ? void 0 : _b.length) {
                                                for (const diagnostic of cachedImportResult.diagnostics) {
                                                    diagnostic.start = node.getStart();
                                                    diagnostic.length = node.getEnd() - node.getStart();
                                                    diagnostics.push(diagnostic);
                                                }
                                            }
                                        }
                                        else {
                                            const deprecatedAliasDiagnostics = isImportDeclarationDeprecated(importDeclaration, checker, log, program, symbol);
                                            if (deprecatedAliasDiagnostics === null || deprecatedAliasDiagnostics === void 0 ? void 0 : deprecatedAliasDiagnostics.length) {
                                                for (const diagnostic of deprecatedAliasDiagnostics) {
                                                    diagnostic.start = node.getStart();
                                                    diagnostic.length = node.getEnd() - node.getStart();
                                                    diagnostics.push(diagnostic);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        ts.forEachChild(node, visit);
                    };
                    visit(sourceFile);
                }
                return priorDiagnostics.concat(diagnostics);
            };
            // Override suggestion diagnostics
            proxy.getSuggestionDiagnostics = (fileName) => {
                const priorDiagnostics = oldGetSuggestionDiagnostics(fileName);
                return checkDiagnostics(fileName, priorDiagnostics, ts.DiagnosticCategory.Suggestion);
            };
            // Override semantic diagnostics
            proxy.getSemanticDiagnostics = (fileName) => {
                const priorDiagnostics = oldGetSemanticDiagnostics(fileName);
                return checkDiagnostics(fileName, priorDiagnostics, ts.DiagnosticCategory.Warning);
            };
            return proxy;
        }
    };
};
module.exports = init;
//# sourceMappingURL=index%20copy.js.map